/*
 * kulock.c - data structures for kernel portion of user level locks
 *
 * Copyright (C) 2001 IBM
 *
 * Written 2001 by Hubertus Franke, IBM Coorporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 02 Sep 2001 - 0.9.0 Initial version
 * 21 Jan 2002 - 1.0.0 Support for kernel callback
 *
 */

/*---------------------------------------------------------------------------
 * This software implements support for fast locking under linux in multi
 * process environments on shared memory architectures. Under
 * fast locking, the lock acquisition/release is split into two parts
 * a fast attempt in user levelwhich is based on 
 * (a) an atomic modification of a lock word placed in a memory 
 *     region shared by processes or threads. If uncontended the lock is
 *     marked busy by the atomic operation and the requesting task can 
 *     continue without entering the kernel
 * (b) in the contested case the kernel routine is called
 *
 *--------------------------------------------------------------------------
 * 
 * Interface to enhanced kernel
 *
 * Kernel hooks are required for clean up: there are 2 fundamental ways to
 * to this 
 *     (a) modify the kernel source code 
 *     (b) hook the vm_ops
 * (b) has some problems on shared memory segments, as the clean up of the
 *     segment is postponed. I provide it here to allow some basic rapid
 *     development and changes and until support for (a) is made available
 *     in the kernel
 * 
 *
 *--------------------------------------------------------------------------*/

//-----------  SETUP

#define VANILLA_KERNEL    // def if kernel was not patched with callback
#define PROVIDE_PROC_FS   // def if a proc filesystem is to be exported
#define AC_STATISTICS     // def if access statistics to be maintained

//-----------  END-SETUP

#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/config.h>
#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/module.h>   /* Specifically, a module */
#include <linux/mm.h>
#include <asm/pgalloc.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <asm/uaccess.h>

#include <linux/ulocks.h>      


/*---------------------------------------------------------------------------
 * Debugging Setup, in case something should go wrong
 *--------------------------------------------------------------------------*/

#define AC_DEBUG(f,x...)   //printk("AC: " f, ##x)
#define SC_DEBUG(f,x...)   //printk("SC: " f, ##x)
#define ACI_DEBUG(f,x...)  //printk("ACI: " f, ##x)
#define KU_DEBUG(f,x...)   //printk("KL: " f, ##x)
#define VMA_DEBUG(f,x...)  //printk("VM: " f, ##x)
#define MOBJ_DEBUG(f,x...) //printk("MO: " f, ##x)
#define HOOK_DEBUG(f,x...) //printk("FC: " f, ##x)
#define INIT_DEBUG(f,x...) printk("INIT: " f , ##x)

/*---------------------------------------------------------------------------
 * Access Statistics 
 *--------------------------------------------------------------------------*/

#if defined(AC_STATISTICS) && defined(PROVIDE_PROC_FS)
#define AC_INC_HIT(idx,num) 						  \
           (atomic_inc((atomic_t*)&access_cache_stats[idx].hits), 	  \
	    atomic_add(num,(atomic_t*)&access_cache_stats[idx].htravers))

#define AC_INC_MISS(idx,num)   						  \
           (atomic_inc((atomic_t*)&access_cache_stats[idx].misses),	  \
	    atomic_add(num,(atomic_t*)&access_cache_stats[idx].mtravers))
#else
#define AC_INC_HIT(idx,num) 
#define AC_INC_MISS(idx,num)
#endif

/*-------------------------------------------------------------------------*/

#define OP_IS_NOT_WAIT_OR_SIGNAL(op)  (((op) & ~1))

struct mobj_t {       /* one per shared file or shared memory segment */
        struct list_head links; 
	unsigned long    numlocks;   /* #locks allocated in this mem-object */
	unsigned long    id;         /* inode backing this memory region    */
	unsigned long    refcount;   /* how many vma's are referring to this*/
        struct list_head kulock_list; /* list of kulocks of this mem-object */
};

typedef struct kulock_t {
	struct semaphore   semaw;        /* writer are blocked here          */
	struct semaphore   semar;        /* readers are blocked here         */
	int                type;         /* type of lock (exclusive,rw,sema) */
	struct mobj_t     *mobj;         /* which memobject it belongs to    */
	unsigned long      ofs;          /* offset in memobject              */
        struct list_head   lookup_links; /* lookup link                      */
	struct list_head   mobj_links;   /* mobj links, if non anonymous     */
} kulock_t;

typedef struct kulockref_t {
	struct kulockref_t *next;   /* link to next ref object */
	struct mm_struct   *mm;     /* address space of ulock */
	void*               vaddr;  /* virtual address of ulock in mm */
	kulock_t           *klock;  /* reference to kern_ulock */
} kulockref_t;


static int vma_to_mobj(struct vm_area_struct *vma, struct mobj_t **mobj_ref);
static void reset_access_cache_stat(void);
static void detach_vma(struct vm_area_struct * vma);

/*---------------------------------------------------------------------------*/
/*  ACCESS-CACHE:                                                            */
/*---------------------------------------------------------------------------*/

#define ACCESS_CACHE_SIZE_BITS 7  // 128
#define ACCESS_CACHE_SIZE (1 << ACCESS_CACHE_SIZE_BITS)

#define GLOBAL_LOOKUP_SIZE_BITS 5 // 32
#define GLOBAL_LOOKUP_SIZE (1 << GLOBAL_LOOKUP_SIZE_BITS)

#define GLOBAL_OBJ_HASH(m,o) \
        (((((unsigned long)(m))+(o)) >> 3) & ((1<< GLOBAL_LOOKUP_SIZE_BITS)-1))

struct ac_stat_t {
	unsigned long     hits;
	unsigned long     misses;
	unsigned long     htravers;
	unsigned long     mtravers;
};

static kulockref_t*       access_cache[ACCESS_CACHE_SIZE];
static struct list_head   kulock_lookup[GLOBAL_LOOKUP_SIZE];
static rwlock_t           ku_lock = RW_LOCK_UNLOCKED;   
static struct ac_stat_t   access_cache_stats[ACCESS_CACHE_SIZE];
static int                num_mobj=0;
static int                num_refs=0;
static int                num_kulocks=0;

static kmem_cache_t*      kulock_cachep     = NULL;
static kmem_cache_t*      kulock_ref_cachep = NULL;


#ifndef USE_VANILLA_KERNEL

/* the VMM provides a callback mechanism when the vma is closed 
 * here we utilize that 
 */
inline
static int vma_ulocks_allowed(struct vm_area_struct *vma)
{
	return (vma->vm_file != NULL);
}

inline
static int vma_ulocks_marked(struct vm_area_struct *vma)
{
	return (vma->vm_flags & VM_SEMA_CB);
}

inline
static void vma_ulocks_mark(struct vm_area_struct *vma)
{
	vma->vm_flags |= VM_SEMA_CB;
	VMA_DEBUG("hooked the vm callback %p %lx\n",vma,vma->vm_flags);
}

static void __kulock_munmap(struct vm_area_struct *vma,
                     	     unsigned long start, unsigned long end)
{
	VMA_DEBUG("kernel-cb MUNMAP %p %lx\n",vma,vma->vm_flags);
	detach_vma(vma);
}

extern void (*ulocks_munmap)(struct vm_area_struct *vma,
                     	     unsigned long start, unsigned long end);

static void hook_kernel(void)
{
	ulocks_munmap = &__kulock_munmap;
}

static void unhook_kernel(void)
{
	ulocks_munmap = &__kulock_munmap;
}

#else

/* the VMM does not provide a callback mechanism when the vma is closed 
 * here we utilize hooking the vm_ops. However due to some coding
 * in the kernel (e.g. shmem.c and shm_vm_ops), stacking transparently
 * the vm_ops is not possible while maintaining the same behavior
 */
static 
void kulock_vm_open(struct vm_area_struct * vma)
{
	HOOK_DEBUG("kulock_vm_open: vma=%p [ %lx-%lx ] ops=%p\n", 
		   vma,vma->vm_start,vma->vm_end,vma->vm_private_data);

	(((struct vm_operations_struct*)vma->vm_private_data)->open)(vma);
}

static
void kulock_vm_close(struct vm_area_struct * vma)
{
	HOOK_DEBUG("kulock_vm_close: vma=%p [ %lx-%lx ] ops=%p\n", 
		   vma,vma->vm_start,vma->vm_end,vma->vm_private_data);
	detach_vma(vma);
	/* unhook the vm_ops */
	vma->vm_ops = (struct vm_operations_struct*)(vma->vm_private_data);
	vma->vm_private_data = NULL;
	if (vma->vm_ops->close) 
		(vma->vm_ops->close)(vma);
	HOOK_DEBUG("kulock_vm_close: vma=%p done ..\n",vma);
}

static
struct page * kulock_vm_nopage(struct vm_area_struct * vma, 
			       unsigned long address, int unused)
{
	HOOK_DEBUG("kulock_nopage: vma=%p [%lx-%lx] ops=%p addr=%lx unused=%d\n", 
		   vma,vma->vm_start,vma->vm_end,vma->vm_private_data,
		   address,unused);
	return (((struct vm_operations_struct*)vma->vm_private_data)->nopage)
		(vma,address,unused);
}

static struct vm_operations_struct kulock_vm_ops = {
	open:   kulock_vm_open,
	close:  kulock_vm_close,
	nopage: kulock_vm_nopage,
};

static int vma_ulocks_allowed(struct vm_area_struct *vma)
{
	return (vma->vm_file != NULL);
}

static int vma_ulocks_marked(struct vm_area_struct *vma)
{
	return (vma->vm_ops == &kulock_vm_ops);
}

static void vma_ulocks_mark(struct vm_area_struct *vma)
{
	if (!vma_ulocks_marked(vma)) {
		VMA_DEBUG("hooked vm_ops vma_flags %lx vm_prot\n",vma->vm_flags);
		vma->vm_private_data = vma->vm_ops;
		vma->vm_ops = &kulock_vm_ops;
	}
}

static void hook_kernel(void)   { }
static void unhook_kernel(void) { }

#endif

/**************************************************************************/

static inline kulock_t *alloc_kulock()
{
	num_kulocks++;
	return ( (kulock_t*) kmem_cache_alloc(kulock_cachep,GFP_KERNEL) );
}

static inline void free_kulock(kulock_t *obj)
{
	num_kulocks--;
	kmem_cache_free(kulock_cachep,obj);
}

static inline kulockref_t *alloc_kulock_ref()
{
	num_refs++;
	return ( (kulockref_t*) kmem_cache_alloc(kulock_ref_cachep,
						      GFP_KERNEL) );
}

static inline void free_kulock_ref(kulockref_t *obj)
{
	num_refs--;
	kmem_cache_free(kulock_ref_cachep,obj);
}

static
void kern_ulock_init(void)
{
	int i;
	kulock_ref_cachep = kmem_cache_create("kern_ulock_ref",
					      sizeof(struct kulockref_t),
					      0,0,NULL,NULL);
	kulock_cachep     = kmem_cache_create("kern_ulock",
					      sizeof(struct kulock_t),
					      0,0,NULL,NULL);
	for(i = 0; i < GLOBAL_LOOKUP_SIZE ; i++) 
		INIT_LIST_HEAD(&kulock_lookup[i]);
	for(i = 0; i < ACCESS_CACHE_SIZE ; i++) 
		access_cache[i] = NULL;
	reset_access_cache_stat();
	hook_kernel();
}

static
int kern_ulock_exit()
{
	int rc = 0;
	unhook_kernel();
	if (kulock_cachep && (num_kulocks == 0))  
		rc += kmem_cache_destroy(kulock_cachep);
	if (kulock_ref_cachep && (num_refs == 0))
		rc += 2*kmem_cache_destroy(kulock_ref_cachep);
	return rc;
}

static inline 
int hash_ulock(struct mm_struct *mm, void* vaddr)
{
	return ( (((((ulong)mm) >> 7) ^ 
		   (((ulong)vaddr) >> 3)) ^ 
		  ((ulong)vaddr >> 7)) & (ACCESS_CACHE_SIZE-1) );
}

static 
void invalidate_access_cache_vma(struct vm_area_struct *vma, int delete_locks)
{
	struct mm_struct *mm = vma->vm_mm;
	void *start   = (void*)vma->vm_start;
	void *end     = (void*)vma->vm_end;

	int i;
	kulockref_t *rptr,*next,**prevnext;

	for ( i=0 ; i<ACCESS_CACHE_SIZE ; i++) {
		if (!access_cache[i]) 
			continue;
		prevnext = &access_cache[i];
		rptr = *prevnext;
		while (rptr) {
			next = rptr->next;
			if ((rptr->mm == mm) && 
			    (rptr->vaddr >= start) && (rptr->vaddr < end))
			{
				AC_DEBUG("    invalidate cache %p %p:%p\n",
					 rptr,rptr->mm,rptr->vaddr);
				if (delete_locks)
					free_kulock(rptr->klock);
				free_kulock_ref(rptr);
				*prevnext = next;
			} else 
				prevnext = &rptr->next;
			rptr = next;
		}
	}
}

static 
int invalidate_access_cache_vaddr(void* vaddr)
{
	struct mm_struct *mm = current->mm;
	int i;
	kulockref_t *rptr,*next,**prevnext;

	for ( i=0 ; i<ACCESS_CACHE_SIZE ; i++) {
		if (!access_cache[i]) 
			continue;
		prevnext = &access_cache[i];
		rptr = *prevnext;
		while (rptr) {
			next = rptr->next;
			if ((rptr->mm == mm) && (rptr->vaddr == vaddr))
			{
				AC_DEBUG("    invalidate cache %p %p:%p\n",
					 rptr,rptr->mm,rptr->vaddr);
				if (list_empty(&rptr->klock->mobj_links))
					free_kulock(rptr->klock);
				free_kulock_ref(rptr);
				return 0;
			} else {
				prevnext = &rptr->next;
				rptr = next;
			}
		}
	}
	return -EINVAL; 
}

static 
void invalidate_access_cache(void)
{
	/* invalidate access cache for non-anonymous objects */
	int i;
	kulockref_t *rptr,*next;

	for ( i=0 ; i<ACCESS_CACHE_SIZE ; i++) {
		next = access_cache[i];
		while ((rptr=next)) {
			next = rptr->next;
			ACI_DEBUG("    invalidate cache %p:%p\n",
				  rptr->mm,rptr->vaddr);
			if (list_empty(&rptr->klock->mobj_links))
				free_kulock(rptr->klock);
			free_kulock_ref(rptr);
		}
		access_cache[i] = NULL;
	}
}

static
int get_kulock(void* vaddr, kulock_t **kulock_ref)
{
	kulockref_t   *rptr;
	kulockref_t   *head_ac;
	kulock_t      *kulock = NULL;
	struct mm_struct  *mm = current->mm;
	struct vm_area_struct   *vma;
	unsigned long boffset=0;
        struct list_head *tmp;     
	unsigned int  ahval;  /* hash value for access cache */
	unsigned int  ghval;  /* hash value for lookup table */
	struct mobj_t *mobj;
	int acnt;
	int rc;

	/* find in the access cache */

	ahval = hash_ulock(mm,vaddr);
	acnt = 0;
	read_lock(&ku_lock);
	head_ac = rptr = access_cache[ahval];
	while (rptr) {
		acnt++;
		if ((rptr->mm == mm) && (rptr->vaddr == vaddr)) {
			read_unlock(&ku_lock);
			AC_INC_HIT(ahval,acnt);
			*kulock_ref = rptr->klock;
			return 0;
		}
		rptr = rptr->next;
	}
	read_unlock(&ku_lock);
	AC_INC_MISS(ahval,acnt);

	/* we couldn't find the lock in the access cache identify the 
	 * global object and then check out the global Ulock map
	 */

	down_write(&mm->mmap_sem);  
	vma = find_vma (mm, (unsigned long) vaddr);
	if (vma == NULL) {
		rc = -EFAULT;
		goto out;
	}
	if (!vma_ulocks_allowed(vma)) {
		rc = -EACCES;
		goto out;
	}

	write_lock(&ku_lock);
	if ((rc = vma_to_mobj(vma,&mobj)))
		goto unlock_and_out;

	ghval = 0;
	if (mobj != NULL) {
		boffset = ((unsigned long)vaddr - vma->vm_start) + 
			(vma->vm_pgoff << PAGE_SHIFT);
		/* find the object in the global object lookup table */
		
		ghval = GLOBAL_OBJ_HASH(mobj,boffset);
		
		list_for_each(tmp, &kulock_lookup[ghval]) {
			kulock_t *p;
			p = list_entry(tmp,kulock_t, lookup_links);
			if ( (p->mobj == mobj) && (p->ofs == boffset) )
			{
				kulock = p;
				break;
			}
		}
	}

	if (kulock == NULL) {
		/* allocate a new object */
		ulock_t ulock;

		if (copy_from_user(&ulock, (void*)vaddr, sizeof(ulock)) ||
		    ((ulock.type < 0) || (ulock.type > 1)))
		{
			rc = -EINVAL;
			goto unlock_and_out;
		}

		kulock = alloc_kulock();
		if (kulock == NULL) {
			rc = -ENOMEM;
			goto unlock_and_out;
		}
		kulock->type     = ulock.type;
		kulock->mobj     = mobj;
		kulock->ofs      = boffset;
		sema_init(&kulock->semaw,0);  /* first down will wait */
		sema_init(&kulock->semar,0);  /* first down will wait */

		if (mobj) {
			list_add(&kulock->lookup_links, &kulock_lookup[ghval]);
			list_add(&kulock->mobj_links, &mobj->kulock_list);
		} else {
			INIT_LIST_HEAD(&kulock->mobj_links);
			INIT_LIST_HEAD(&kulock->lookup_links);
		}
	}

	/* create access cache entry, but verify that a race condition
	 * did not result in somebody else already creating the entry.
	 * duplicate kulocks can not be created by this race condition
	 * only duplicate access cache entries.
	 */
	if (head_ac != access_cache[ahval]) {
		rptr = access_cache[ahval];
		while (rptr) {
			if ((rptr->mm == mm) && (rptr->vaddr == vaddr)) {
				kulock = rptr->klock;
				goto unlock_and_out;
			}
			rptr = rptr->next;
		}
	}
	rptr        = alloc_kulock_ref();
	if (rptr == NULL) {
		rc = -ENOMEM;
		goto unlock_and_out;
	}
	rptr->mm    = mm;
	rptr->vaddr = vaddr;
	rptr->klock = kulock;
	rptr->next  = access_cache[ahval];
	access_cache[ahval] = rptr;
	KU_DEBUG("vaddr=%p mm=%p vma=%p %lx [%lx-%lx] bo=%lx:%lx h=%d g=%d\n", 
		 vaddr,mm,vma,vma->vm_flags,vma->vm_start,vma->vm_end,
		 mobj?mobj->id:0,boffset,ahval,ghval);

 unlock_and_out:	
	write_unlock(&ku_lock);
 out:
	up_write(&mm->mmap_sem);
	*kulock_ref = kulock;
	return rc;
}
	
/*---------------------------------------------------------------------------*
 *                maintaining the global <mobj> state                        *
 *---------------------------------------------------------------------------*/

LIST_HEAD(mobjs);                      /* keeps all memory objects in a list */

static
void delete_mobj_kulocks(struct mobj_t *mobj)
{
	struct list_head *tmp,*next;
        list_for_each_safe(tmp, next, &mobj->kulock_list) {
		kulock_t *p;
                p = list_entry(tmp,kulock_t, mobj_links);
		KU_DEBUG("  deleting kulock=%p\n",p);
		list_del(&p->mobj_links);   // detach from mobj
		list_del(&p->lookup_links); // remove from lookup table
		free_kulock(p);
	}
}

static
int vma_to_mobj(struct vm_area_struct *vma, struct mobj_t **mobj_ref)
{
	unsigned long bid;
	struct list_head *tmp;
	struct mobj_t *mobj;

	if (vma->vm_file == NULL) {
		if (!vma_ulocks_marked(vma)) 
			vma_ulocks_mark(vma);
		*mobj_ref = NULL;
		return 0;
	}
	bid     = vma->vm_file->f_dentry->d_inode->i_ino;
	
	/* increment the reference count of an MOBJ 
	 * if the MOBJ does not exist, then create and initialize it 
	 */
	
        list_for_each(tmp, &mobjs) {
                mobj = list_entry(tmp,struct mobj_t, links);
		if (mobj->id == bid)
		{
			MOBJ_DEBUG("found %p <%lx:%ld> new_vma=%d\n",
				   mobj,bid,mobj->refcount,!vma_ulocks_marked(vma));
			goto out;
		}
        }
	mobj = kmalloc(sizeof(struct mobj_t),GFP_KERNEL);
	if (mobj == NULL) {
		*mobj_ref = NULL;
		return -ENOMEM;
	}
	mobj->numlocks = 0;
	mobj->refcount = 0;
	mobj->id       = bid;
	INIT_LIST_HEAD(&mobj->kulock_list);
	list_add(&mobj->links, &mobjs);
	num_mobj++;
	MOBJ_DEBUG("new %p <%lx:1>\n",mobj,bid);
 out:
	if (!vma_ulocks_marked(vma)) {
		mobj->refcount++;
		vma_ulocks_mark(vma);
	}
	*mobj_ref = mobj;
	return 0;
}	

static
void detach_vma(struct vm_area_struct * vma)
{
	struct mobj_t  *mobj;

	/* executed under vma->vm_mm->mmap_sem semaphore */

	write_lock(&ku_lock);
	if (!vma_to_mobj(vma,&mobj)) {
		if (mobj) {
			VMA_DEBUG("remove vma=%p <-:%lx:%lx> memobj=<%ld:%ld>\n",
				  vma,vma->vm_start,vma->vm_end,
				  mobj->id,mobj->refcount-1);
			invalidate_access_cache_vma (vma,0);
			if (--mobj->refcount == 0) {
				MOBJ_DEBUG("Deleting MOBJ <%lx>\n",mobj->id);
				list_del(&mobj->links);  
				delete_mobj_kulocks(mobj);
				kfree(mobj);
				num_mobj--;
			}
		} else {
			VMA_DEBUG("remove anonymous vma=%p <-:%lx:%lx>\n",
				  vma,vma->vm_start,vma->vm_end);
			invalidate_access_cache_vma(vma,1);
		}
	}
	write_unlock(&ku_lock);
}


int sys_ulock_op (int op, void* ulock, long rwvalue )
{
	kulock_t *klock;
	int rc;

	if (OP_IS_NOT_WAIT_OR_SIGNAL(op))
		goto not_fast_op;

	SC_DEBUG("   fastop(%d,%p,%ld)\n",op,ulock,rwvalue);

	/* here we only deal with ULOCK_OP_WAIT and ULOCK_OP_SIGNAL */
	if ((rc = get_kulock(ulock,&klock))) 
		goto fixup;

	switch (klock->type) {

	case ULOCK_SEMAPHORE: 
		if (op == ULOCK_OP_WAIT) { 
			if (down_interruptible(&klock->semaw))
				return -ERESTARTSYS;
		} else {
			up(&klock->semaw); 
		}
		break;
	case ULOCK_RWLOCK: 
		if (op == ULOCK_OP_WAIT) {
			if (rwvalue) {
				if (down_interruptible(&klock->semar))
					return -ERESTARTSYS;
			} else {
				if (down_interruptible(&klock->semaw))
					return -ERESTARTSYS;
			}
		} else {
			if ((rwvalue < -1) || (rwvalue > 100))
				return -EINVAL;
			if (rwvalue == -1) {
				up(&klock->semaw);
			} else  {
				int i;
				for ( i=0 ; i<rwvalue ; i++)
					up(&klock->semar);
				break;
			}
		}
	}
	return 0; 
	
	/* here we only deal with operations other than
	 * ULOCK_OP_WAIT and ULOCK_OP_SIGNAL 
	 */
 not_fast_op:

	SC_DEBUG("   slowop(%d,%p,%ld)\n",op,ulock,rwvalue);
	switch (op) {
	case ULOCK_OP_CREATE:
		return get_kulock(ulock,&klock);
	case ULOCK_OP_DESTROY:
		return invalidate_access_cache_vaddr(ulock);
	default:
		return -ENOSYS;
	}

 fixup:
	/* an error occured before we got to the lock operation 
	 * the user mode status word is not in sync with semaphores
	 * in the kernel.
	 * There are two types of errors
	 * (i) the ones that could be fixed
	 *
	 *    ENOMEM: Insufficient kernel memory at this point
	 *
	 * (ii) the ones we can forget about
         *    EFAULT: illegal address was specified
	 *            system call was issued directly without the 
         *            the wrapper functions, otherwise the 
	 *            thw wrapper functions would have SEGV
	 *    EINVAL: lock type unknown
	 *    EACCES: MAP_SEMAPHORE not specified on VMA
	 */
	
	/* at this point it doesn't really make sense to fix it */
	return rc;
	
}

void reset_access_cache_stat(void)
{
	int i;

	for(i = 0; i < ACCESS_CACHE_SIZE ; i++) {
		access_cache_stats[i].hits = 0;
		access_cache_stats[i].misses = 0;
		access_cache_stats[i].htravers = 0;
		access_cache_stats[i].mtravers = 0;
	}
}

#ifdef PROVIDE_PROC_FS

#include <linux/proc_fs.h>

int kulock_read_proc(char *page, char **start, off_t off, 
                        int count, int *eof, void *data)
{
	int i;
        int len = 0;
	struct ac_stat_t total = { 0, 0 , 0 , 0 };
	unsigned long htrav, mtrav;

	for (i=0;i<ACCESS_CACHE_SIZE;i++) {
		total.hits += access_cache_stats[i].hits;
		total.misses += access_cache_stats[i].misses;
		total.htravers += access_cache_stats[i].htravers;
		total.mtravers += access_cache_stats[i].mtravers;
	}
	htrav = total.hits   ? (100*total.htravers)/total.hits   : 0;
	mtrav = total.misses ? (100*total.mtravers)/total.misses : 0;

	len =  sprintf(page,"#kulocks=%d #refs=%d #mobj=%d\n",
		       num_kulocks,num_refs,num_mobj);
	len += sprintf(page+len,"hits=%ld htravers=%ld (%ld.%02ld) ",
		       total.hits, total.htravers, htrav / 100, htrav % 100);
	len += sprintf(page+len,"misses=%ld mtravers=%ld (%ld.%02ld)\n",
		      total.misses, total.mtravers, mtrav / 100, mtrav % 100);
        if (len <= off+count) *eof = 1;
        *start = page + off;
        len -= off;
        if (len>count) len = count;
        if (len<0) len = 0;
        return len;
}

int kulock_write_proc (struct file *file, const char *buffer,
			  unsigned long count, void *data)
{
        char kbuf[256];
	char *cptr = kbuf;
        int bytes = (count > 256) ? 256 : count;

        if (__generic_copy_from_user(&kbuf,buffer,bytes))
                return EINVAL;

	write_lock(&ku_lock);
	while (*cptr && cptr < &kbuf[256]) {
		switch (*cptr) {
		case 'i':
			invalidate_access_cache();
			break;
		case 's':
			reset_access_cache_stat();
			break;
		}
		cptr++;
	}
	write_unlock(&ku_lock);
        return count;
}

static 
void create_proc_fs(void)
{
	struct proc_dir_entry *res;
        res = create_proc_entry("kulocks", S_IWUSR | S_IRUGO, &proc_root);
        if (res) {
                res->read_proc  = kulock_read_proc;
                res->write_proc = kulock_write_proc;
                res->data = NULL;
        }
}

static inline 
void remove_proc_fs(void)
{
        remove_proc_entry("kulocks", &proc_root);
}

#else
inline static create_proc_fs(void) { }
inline static remove_proc_fs(void) { }
#endif

/****************************************************************************
 *  MODULE support
 ****************************************************************************/

#include <linux/proc_fs.h>

static int syscall_num = 232;
extern int sys_call_table[];
   
int prev_sys_call_entry;
    
int init_module()
{
	INIT_DEBUG("ULocks initializing\n");
	prev_sys_call_entry = sys_call_table[syscall_num];
	sys_call_table[syscall_num] = (int) sys_ulock_op;
	kern_ulock_init();
	create_proc_fs();
	INIT_DEBUG("ULocks initialized\n");
        return 0;
}
 
void cleanup_module()
{
	int rc;
	sys_call_table[syscall_num] = prev_sys_call_entry;

	rc = kern_ulock_exit();
	INIT_DEBUG("ULocks removed  rc=%d\n",rc);
}



